# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Git dataset provider."""

import glob
import os
import shutil
from collections import defaultdict
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Union

from renku.core import errors
from renku.core.dataset.pointer_file import create_external_file
from renku.core.dataset.providers.api import AddProviderInterface, ProviderApi, ProviderPriority
from renku.core.storage import pull_paths_from_storage
from renku.core.util import communication
from renku.core.util.git import clone_repository, get_cache_directory_for_repository
from renku.core.util.metadata import is_linked_file
from renku.core.util.os import delete_dataset_file, get_files, is_subpath
from renku.core.util.urls import check_url, remove_credentials
from renku.domain_model.dataset import RemoteEntity
from renku.domain_model.project_context import project_context
from renku.infrastructure.immutable import DynamicProxy

if TYPE_CHECKING:
    from renku.core.dataset.providers.models import DatasetAddMetadata, DatasetUpdateMetadata, ProviderParameter


class GitProvider(ProviderApi, AddProviderInterface):
    """Git provider."""

    priority = ProviderPriority.NORMAL
    name = "Git"
    is_remote = True

    @staticmethod
    def supports(uri: str) -> bool:
        """Whether or not this provider supports a given URI."""
        is_remote, is_git = check_url(uri)
        return is_remote and is_git

    @staticmethod
    def get_add_parameters() -> List["ProviderParameter"]:
        """Returns parameters that can be set for add."""
        from renku.core.dataset.providers.models import ProviderParameter

        return [
            ProviderParameter(
                "sources",
                flags=["s", "source"],
                default=None,
                help="Path(s) within remote git repo to be added",
                multiple=True,
            ),
            ProviderParameter(
                "revision",
                flags=["r", "ref"],
                default=None,
                help="Add files from a specific commit/tag/branch.",
                type=str,
            ),
        ]

    def get_metadata(
        self,
        uri: str,
        destination: Path,
        *,
        sources: Optional[List[Union[Path, str]]] = None,
        revision: Optional[str] = None,
        **kwargs,
    ) -> List["DatasetAddMetadata"]:
        """Get metadata of files that will be added to a dataset."""
        from renku.core.dataset.providers.models import DatasetAddAction, DatasetAddMetadata

        destination_exists = destination.exists()
        destination_is_dir = destination.is_dir()

        remote_repository = clone_repository(
            url=uri,
            path=get_cache_directory_for_repository(url=uri),
            checkout_revision=revision,
            depth=None,
            clean=True,
        )

        def check_sources_are_within_remote_repo():
            if not sources:
                return
            for source in sources:
                if not is_subpath(path=remote_repository.path / source, base=remote_repository.path):
                    raise errors.ParameterError(f"Path '{source}' is not within the repository")

        def get_source_paths() -> Set[Path]:
            """Return all paths from the repo that match a source pattern."""
            if not sources:
                return set(remote_repository.path.glob("*"))  # type: ignore

            paths = set()
            for source in sources:
                # NOTE: Normalized source to resolve .. references (if any). This preserves wildcards.
                normalized_source = os.path.normpath(source)
                absolute_source = os.path.join(remote_repository.path, normalized_source)  # type: ignore
                # NOTE: Path.glob("root/**") does not return correct results (e.g. it include ``root`` in the result)
                subpaths = {Path(p) for p in glob.glob(absolute_source)}
                if len(subpaths) == 0:
                    raise errors.ParameterError("No such file or directory", param_hint=str(source))
                paths |= subpaths

            return paths

        def should_copy(source_paths: List[Path]) -> bool:
            n_paths = len(source_paths)
            has_multiple_sources = n_paths > 1
            source_is_dir = has_multiple_sources or (n_paths == 1 and source_paths[0].is_dir())

            if source_is_dir and destination_exists and not destination_is_dir:
                raise errors.ParameterError(f"Destination is not a directory: '{destination}'")

            return has_multiple_sources or (destination_exists and destination_is_dir)

        def get_file_metadata(src: Path, dst: Path) -> Optional["DatasetAddMetadata"]:
            path_in_src_repo = src.relative_to(remote_repository.path)  # type: ignore
            path_in_dst_repo = dst.relative_to(project_context.path)

            already_copied = path_in_dst_repo in new_files  # A path with the same destination is already copied
            new_files[path_in_dst_repo].append(path_in_src_repo)
            if already_copied:
                return None

            checksum = remote_repository.get_object_hash(revision="HEAD", path=path_in_src_repo)  # type: ignore
            if not checksum:
                raise errors.FileNotFound(f"Cannot find '{file}' in the remote project")

            return DatasetAddMetadata(
                entity_path=path_in_dst_repo,
                url=remove_credentials(uri),
                based_on=RemoteEntity(checksum=checksum, path=path_in_src_repo, url=uri),
                action=DatasetAddAction.MOVE,
                source=src,
                destination=dst,
            )

        check_sources_are_within_remote_repo()

        results = []
        new_files: Dict[Path, List[Path]] = defaultdict(list)

        paths = get_source_paths()
        with project_context.with_path(remote_repository.path):
            pull_paths_from_storage(project_context.repository, *paths)
        is_copy = should_copy(list(paths))

        for path in paths:
            dst_root = destination / path.name if is_copy else destination

            for file in get_files(path):
                src = file
                relative_path = file.relative_to(path)
                dst = dst_root / relative_path

                metadata = get_file_metadata(src, dst)
                if metadata:
                    results.append(metadata)

        duplicates = [v for v in new_files.values() if len(v) > 1]
        if duplicates:
            files = {str(p) for paths in duplicates for p in paths}
            files_str = "/n/t".join(sorted(files))
            communication.warn(f"The following files overwrite each other in the destination project:/n/t{files_str}")

        return results

    def update_files(
        self,
        files: List[DynamicProxy],
        dry_run: bool,
        delete: bool,
        context: Dict[str, Any],
        ref: Optional[str] = None,
        **kwargs,
    ) -> List["DatasetUpdateMetadata"]:
        """Update dataset files from the remote provider."""
        from renku.core.dataset.providers.models import DatasetUpdateAction, DatasetUpdateMetadata

        if "visited_repos" not in context:
            context["visited_repos"] = {}

        progress_text = "Checking git files for updates"

        results: List[DatasetUpdateMetadata] = []

        try:
            communication.start_progress(progress_text, len(files))
            for file in files:
                communication.update_progress(progress_text, 1)
                if not file.based_on:
                    continue

                based_on = file.based_on
                url = based_on.url
                if url in context["visited_repos"]:
                    remote_repository = context["visited_repos"][url]
                else:
                    communication.echo(msg="Cloning remote repository...")
                    path = get_cache_directory_for_repository(url=url)
                    remote_repository = clone_repository(url=url, path=path, checkout_revision=ref)
                    context["visited_repos"][url] = remote_repository

                checksum = remote_repository.get_object_hash(path=based_on.path, revision="HEAD")
                found = checksum is not None
                changed = found and based_on.checksum != checksum

                src = remote_repository.path / based_on.path
                dst = project_context.metadata_path.parent / file.entity.path

                if not found:
                    if not dry_run and delete:
                        delete_dataset_file(dst, follow_symlinks=True)
                        project_context.repository.add(dst, force=True)
                    results.append(DatasetUpdateMetadata(entity=file, action=DatasetUpdateAction.DELETE))
                elif changed:
                    if not dry_run:
                        # Fetch file if it is tracked by Git LFS
                        pull_paths_from_storage(remote_repository, remote_repository.path / based_on.path)
                        if is_linked_file(path=src, project_path=remote_repository.path):
                            delete_dataset_file(dst, follow_symlinks=True)
                            create_external_file(target=src.resolve(), path=dst)
                        else:
                            shutil.copy(src, dst)
                        file.based_on = RemoteEntity(
                            checksum=checksum, path=based_on.path, url=based_on.url  # type: ignore
                        )
                    results.append(DatasetUpdateMetadata(entity=file, action=DatasetUpdateAction.UPDATE))
        finally:
            communication.finalize_progress(progress_text)

        return results



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Migration utility functions."""

import os
import posixpath
import threading
import uuid
from typing import Optional, cast
from urllib.parse import ParseResult, quote, urljoin, urlparse

from renku.core.util.yaml import read_yaml
from renku.domain_model.project_context import project_context

OLD_METADATA_PATH = "metadata.yml"
OLD_DATASETS_PATH = "datasets"
OLD_WORKFLOW_PATH = "workflow"

thread_local_storage = threading.local()


def generate_url_id(url_str, url_id):
    """Generate @id field for Url."""
    url = url_str or url_id
    if url:
        parsed_result = urlparse(url)
        id_ = ParseResult("", *parsed_result[1:]).geturl()
    else:
        id_ = str(uuid.uuid4())

    host = "localhost"
    if project_context.has_context():
        host = project_context.remote.host or host
    host = os.environ.get("RENKU_DOMAIN") or host

    return urljoin(f"https://{host}", posixpath.join("/urls", quote(id_, safe="")))


def generate_dataset_tag_id(name, commit):
    """Generate @id field for DatasetTag."""
    host = "localhost"
    if project_context.has_context():
        host = project_context.remote.host or host
    host = os.environ.get("RENKU_DOMAIN") or host

    name = f"{name}@{commit}"

    return urljoin(f"https://{host}", posixpath.join("/dataset-tags", quote(name, safe="")))


def generate_dataset_id(identifier):
    """Generate @id field."""
    # Determine the hostname for the resource URIs.
    # If RENKU_DOMAIN is set, it overrides the host from remote.
    # Default is localhost.
    host = "localhost"
    if project_context.has_context():
        host = project_context.remote.host or host
    host = os.environ.get("RENKU_DOMAIN") or host

    # always set the id by the identifier
    return urljoin(f"https://{host}", posixpath.join("/datasets", quote(identifier, safe="")))


def generate_dataset_file_url(filepath):
    """Generate url for DatasetFile."""
    try:
        project = project_context.project
        if not project:
            return
    except ValueError:
        from renku.core.migration.models.v9 import Project

        metadata_path = project_context.metadata_path.joinpath(OLD_METADATA_PATH)
        project = Project.from_yaml(metadata_path)

        project_id = urlparse(project._id)
    else:
        project_id = urlparse(project.id)

    filepath = quote(filepath, safe="/")
    path = posixpath.join(project_id.path, "files", "blob", filepath)
    project_id = project_id._replace(path=path)

    return project_id.geturl()


def migrate_types(data):
    """Fix data types."""
    type_mapping = {
        "dcterms:creator": ["prov:Person", "schema:Person"],
        "schema:Person": ["prov:Person", "schema:Person"],
        str(sorted(["foaf:Project", "prov:Location"])): ["prov:Location", "schema:Project"],
        "schema:DigitalDocument": ["prov:Entity", "schema:DigitalDocument", "wfprov:Artifact"],
    }

    def remove_type(data_):
        data_.pop("@type", None)
        for key, value in data_.items():
            if isinstance(value, dict):
                remove_type(value)
            elif isinstance(value, (list, tuple, set)):
                for v in value:
                    if isinstance(v, dict):
                        remove_type(v)

    def replace_types(data_):
        for key, value in data_.items():
            if key == "@context" and isinstance(value, dict):
                remove_type(value)

            if key == "@type":
                if not isinstance(value, str):
                    value = str(sorted(value))
                new_type = type_mapping.get(value)
                if new_type:
                    data_[key] = new_type
            elif isinstance(value, dict):
                replace_types(value)
            elif isinstance(value, (list, tuple, set)):
                for v in value:
                    if isinstance(v, dict):
                        replace_types(v)

    replace_types(data)

    return data


def get_pre_0_3_4_datasets_metadata():
    """Return paths of dataset metadata for pre 0.3.4."""
    from renku.core.constant import DEFAULT_DATA_DIR as DATA_DIR

    project_is_pre_0_3 = int(read_project_version()) < 2
    if project_is_pre_0_3:
        return (project_context.path / DATA_DIR).glob(f"*/{OLD_METADATA_PATH}")
    return []


def read_project_version() -> str:
    """Read project version from metadata file."""
    try:
        return project_context.project.version
    except (NotImplementedError, ValueError):
        metadata_path = project_context.metadata_path / OLD_METADATA_PATH
        if not os.path.exists(metadata_path):
            return "1"

        yaml_data = read_yaml(metadata_path)
        return read_project_version_from_yaml(yaml_data)


def read_latest_agent() -> Optional[str]:
    """Read project version from metadata file."""
    import pyld

    try:
        return project_context.latest_agent
    except (NotImplementedError, ValueError):
        metadata_path = project_context.metadata_path / OLD_METADATA_PATH
        if not os.path.exists(metadata_path):
            raise

        yaml_data = read_yaml(metadata_path)
        jsonld = pyld.jsonld.expand(yaml_data)[0]
        jsonld = normalize(jsonld)
        return cast(str, _get_jsonld_property(jsonld, "http://schema.org/agent", "pre-0.11.0"))


def read_project_version_from_yaml(yaml_data) -> str:
    """Read project version from YAML data."""
    import pyld

    jsonld = pyld.jsonld.expand(yaml_data)[0]
    jsonld = normalize(jsonld)
    return cast(str, _get_jsonld_property(jsonld, "http://schema.org/schemaVersion", "1"))


def _get_jsonld_property(jsonld, property_name, default=None):
    """Return property value from expanded JSON-LD data."""
    value = jsonld.get(property_name)
    if not value:
        return default
    if isinstance(value, list) and len(value) == 1 and isinstance(value[0], dict) and "@value" in value[0]:
        value = value[0]["@value"]
    return value


def normalize(value):
    """Normalize an expanded JSON-LD."""
    if isinstance(value, list):
        if len(value) == 1 and isinstance(value[0], dict) and "@value" in value[0]:
            return value[0]["@value"]
        return [normalize(v) for v in value]

    if isinstance(value, dict):
        if "@value" in value:
            return value["@value"]
        else:
            return {k: normalize(v) for k, v in value.items()}

    return value


def get_datasets_path():
    """Get the old datasets metadata path."""
    return getattr(thread_local_storage, "temporary_datasets_path", project_context.metadata_path / OLD_DATASETS_PATH)


def set_temporary_datasets_path(temporary_datasets_path):
    """Set a temporary datasets metadata path."""
    thread_local_storage.temporary_datasets_path = temporary_datasets_path


def is_using_temporary_datasets_path():
    """Check if temporary datasets path is set."""
    return bool(getattr(thread_local_storage, "temporary_datasets_path", None))


def unset_temporary_datasets_path():
    """Unset the current temporary datasets metadata path."""
    if getattr(thread_local_storage, "temporary_datasets_path", None):
        del thread_local_storage.temporary_datasets_path



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Migrate project to the latest Renku version.

Description
~~~~~~~~~~~

Bring the project up to date with the current version of Renku.
This does not usually affect how you use Renku and no data is lost.

Commands and options
~~~~~~~~~~~~~~~~~~~~

.. rst-class:: cli-reference-commands

.. click:: renku.ui.cli.migrate:migrate
   :prog: renku migrate
   :nested: full

Dockerfile and template
~~~~~~~~~~~~~~~~~~~~~~~

In addition, ``renku migrate`` will update your ``Dockerfile` to install the
latest version of ``renku-python``, if supported, making sure your renku
version is up to date in interactive environments as well.

If you created your repository from a project template and the template has
changed since you created the project, it will also update files with their
newest version from the template, without overwriting local changes if there
are any.

You can check if a migration is necessary and what migrations are available
by running

.. code-block:: console

    $ renku migrate -c

.. cheatsheet::
   :group: Misc
   :command: $ renku migrate
   :description: Migrate old metadata to the current Renku version.
   :target: rp

"""
import json
import os
from dataclasses import asdict

import click

import renku.ui.cli.utils.color as color
from renku.command.util import ERROR, INFO
from renku.core.errors import MigrationRequired, ProjectNotSupported
from renku.ui.cli.utils.callback import ClickCallback


@click.command()
@click.option("-c", "--check", is_flag=True, help="Check if migration is required and quit.")
@click.option("-t", "--skip-template-update", is_flag=True, hidden=True, help="Do not update project template files.")
@click.option(
    "-d", "--skip-docker-update", is_flag=True, hidden=True, help="Do not update Dockerfile to current renku version."
)
@click.option("-s", "--strict", is_flag=True, hidden=True, help="Abort migrations if an error is raised.")
@click.option(
    "--preserve-identifiers",
    is_flag=True,
    hidden=True,
    help="Keep original dataset identifier when KG migrates a project.",
)
def migrate(check, skip_template_update, skip_docker_update, strict, preserve_identifiers):
    """Check for migration and migrate to the latest Renku project version."""
    from renku.command.migrate import (
        AUTOMATED_TEMPLATE_UPDATE_SUPPORTED,
        DOCKERFILE_UPDATE_POSSIBLE,
        MIGRATION_REQUIRED,
        NON_RENKU_REPOSITORY,
        TEMPLATE_UPDATE_POSSIBLE,
        UNSUPPORTED_PROJECT,
        check_project,
        migrate_project_command,
    )

    status = check_project().build().execute().output

    template_update_possible = status & TEMPLATE_UPDATE_POSSIBLE and status & AUTOMATED_TEMPLATE_UPDATE_SUPPORTED
    docker_update_possible = status & DOCKERFILE_UPDATE_POSSIBLE

    if check:
        if template_update_possible:
            click.secho(
                INFO
                + "The project template used to create this project has updates which can be applied "
                + "using 'renku migrate'."
            )
        if docker_update_possible:
            click.secho(
                INFO
                + "The Dockerfile refers to an older version of renku and can be updated "
                + "using 'renku migrate'."
            )

        if status & MIGRATION_REQUIRED:
            raise MigrationRequired

    if status & UNSUPPORTED_PROJECT:
        raise ProjectNotSupported
    elif status & NON_RENKU_REPOSITORY:
        click.secho(ERROR + "Not a renku project.")
        return

    if check:
        return

    skip_docker_update = skip_docker_update or not docker_update_possible
    skip_template_update = skip_template_update or not template_update_possible

    communicator = ClickCallback()

    command = migrate_project_command().with_communicator(communicator).with_commit()
    result = command.build().execute(
        skip_template_update=skip_template_update,
        skip_docker_update=skip_docker_update,
        strict=strict,
        preserve_identifiers=preserve_identifiers,
    )

    result, _, _ = result.output

    if result:
        click.secho("OK", fg=color.GREEN)
    else:
        click.secho("No migrations required.")


@click.command(hidden=True)
def migrationscheck():
    """Check status of the project and current renku-python version."""
    from renku.command.migrate import migrations_check

    result = migrations_check().lock_project().build().execute().output
    result_dict = asdict(result)

    if result_dict.get("errors"):
        for key, value in result_dict["errors"]:
            result_dict["errors"][key] = str(value)
    click.echo(json.dumps(result_dict))


@click.command(hidden=True)
@click.argument("paths", type=click.Path(exists=True, dir_okay=True), nargs=-1, required=True)
def check_immutable_template_files(paths):
    """Check specified paths if they are marked immutable in the template."""
    from renku.command.migrate import check_immutable_template_files_command

    result = check_immutable_template_files_command().build().execute(paths=paths)
    paths = result.output
    if paths:
        click.echo(os.linesep.join(paths))
        exit(1)



#  Copyright Swiss Data Science Center (SDSC). A partnership between
#  École Polytechnique Fédérale de Lausanne (EPFL) and
#  Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Docker based interactive session provider."""

import pty
import urllib
import webbrowser
from datetime import datetime
from pathlib import Path
from time import monotonic, sleep
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

from renku.core import errors
from renku.core.config import get_value
from renku.core.constant import ProviderPriority
from renku.core.login import read_renku_token
from renku.core.plugin import hookimpl
from renku.core.session.utils import get_renku_project_name, get_renku_url
from renku.core.util import communication, requests
from renku.core.util.git import get_remote
from renku.core.util.jwt import is_token_expired
from renku.core.util.ssh import SystemSSHConfig
from renku.domain_model.project_context import project_context
from renku.domain_model.session import IHibernatingSessionProvider, Session, SessionStopStatus

if TYPE_CHECKING:
    from renku.core.dataset.providers.models import ProviderParameter


class RenkulabSessionProvider(IHibernatingSessionProvider):
    """A session provider that uses the notebook service API to launch sessions."""

    DEFAULT_TIMEOUT_SECONDS = 300
    # NOTE: Give the renkulab provider the lowest priority so that it's checked last
    priority: ProviderPriority = ProviderPriority.LOWEST

    def __init__(self):
        self.__renku_url: Optional[str] = None
        self.__notebooks_url: Optional[str] = None
        self._force_build: bool = False

    def _renku_url(self) -> str:
        """Get the URL of the renku instance."""
        if not self.__renku_url:
            renku_url = get_renku_url()
            if not renku_url:
                raise errors.RenkulabSessionGetUrlError()
            self.__renku_url = renku_url
        return self.__renku_url

    def _notebooks_url(self) -> str:
        """Get the url of the notebooks API."""
        if not self.__notebooks_url:
            url = urllib.parse.urljoin(self._renku_url(), "api/notebooks")
            self.__notebooks_url = url
        return self.__notebooks_url

    def _get_token(self) -> str:
        """Get the JWT token used to authenticate against Renku."""
        token = read_renku_token(endpoint=self._renku_url())
        if token is None:
            raise errors.AuthenticationError("Please run the renku login command to authenticate with Renku.")
        elif is_token_expired(token):
            raise errors.AuthenticationError(
                "Authentication token is expired: Please run the renku login command to authenticate with Renku."
            )
        return token

    def _auth_header(self) -> Dict[str, str]:
        """Get the authentication header with the JWT token or cookie needed to authenticate with Renku."""
        return {"Authorization": f"Bearer {self._get_token()}"}

    @staticmethod
    def _get_renku_project_name_parts():
        repository = project_context.repository
        if project_context.remote.name and project_context.remote.owner:
            if get_remote(repository, name="renku-backup-origin") and project_context.remote.owner.startswith("repos/"):
                owner = project_context.remote.owner.replace("repos/", "", 1)
            else:
                owner = project_context.remote.owner
            return {
                "namespace": owner,
                "project": project_context.remote.name,
            }
        else:
            # INFO: In this case the owner/name split is not available. The project name is then
            # derived from the combined name of the remote and has to be split up in the two parts.
            parts = get_renku_project_name().split("/")
            return {
                "namespace": "/".join(parts[:-1]),
                "project": parts[:-1],
            }

    def _wait_for_session_status(
        self,
        name: Optional[str],
        status: str,
    ):
        if not name:
            return
        start = monotonic()
        while monotonic() - start < self.DEFAULT_TIMEOUT_SECONDS:
            res = self._send_renku_request(
                "get", f"{self._notebooks_url()}/servers/{name}", headers=self._auth_header()
            )
            if res.status_code == 404 and status == "stopping":
                return
            if res.status_code in [200, 204] and status != "stopping":
                if res.json().get("status", {}).get("state") == status:
                    return
            sleep(5)
        raise errors.RenkulabSessionError(f"Waiting for the session {name} to reach status {status} timed out.")

    def _wait_for_image(
        self,
        image_name: str,
        config: Optional[Dict[str, Any]],
    ):
        """Check if an image exists, and if it does not wait for it to appear.

        Timeout after a specific period of time.
        """
        start = monotonic()
        while monotonic() - start < self.DEFAULT_TIMEOUT_SECONDS:
            if self.find_image(image_name, config):
                return
            sleep(5)
        raise errors.RenkulabSessionError(
            f"Waiting for the image {image_name} to be built timed out."
            "Are you sure that the image was successfully built? This could be the result "
            "of problems with your Dockerfile."
        )

    def pre_start_checks(self, ssh: bool = False, **kwargs):
        """Check if the state of the repository is as expected before starting a session."""
        from renku.core.session.session import ssh_setup

        repository = project_context.repository

        if repository.is_dirty():
            communication.confirm(
                "You have new uncommitted or untracked changes to your repository. "
                "Renku can automatically commit these changes so that it builds "
                "the correct environment for your session. Do you wish to proceed?",
                abort=True,
            )
            repository.add(all=True)
            repository.commit("Automated commit by Renku CLI.")

        if ssh:
            system_config = SystemSSHConfig()

            if not system_config.is_configured:
                if communication.confirm(
                    "Your system is not set up for SSH connections to Renkulab. Would you like to set it up?"
                ):
                    ssh_setup()
                    self._force_build = True
                else:
                    raise errors.RenkulabSessionError(
                        "Can't run ssh session without setting up Renku SSH support. Run without '--ssh' or "
                        "run 'renku session ssh-setup'."
                    )

            if system_config.setup_session_keys():
                self._force_build = True

    def _cleanup_ssh_connection_configs(
        self, project_name: str, running_sessions: Optional[List[Session]] = None
    ) -> None:
        """Cleanup leftover SSH connections that aren't valid anymore.

        Args:
            project_name(str): Name of the project.
            running_sessions(List[Session], optional): List of running sessions to check against, otherwise will be
                gotten from the server.
        """
        if not running_sessions:
            running_sessions = self.session_list(project_name="", ssh_garbage_collection=False)

        system_config = SystemSSHConfig()

        name = self._project_name_from_full_project_name(project_name)

        session_config_paths = [system_config.session_config_path(name, s.id) for s in running_sessions]

        for path in system_config.renku_ssh_root.glob(f"00-{name}*.conf"):
            if path not in session_config_paths:
                path.unlink()

    @staticmethod
    def _remote_head_hexsha():
        remote = get_remote(repository=project_context.repository)

        if remote is None:
            raise errors.GitRemoteNotFoundError()

        return remote.head

    def _send_renku_request(self, verb: str, *args, **kwargs):
        response = getattr(requests, verb)(*args, **kwargs)
        if response.status_code == 401:
            # NOTE: Check if logged in to KC but not the Renku UI
            token = read_renku_token(endpoint=self._renku_url())
            if token and not is_token_expired(token):
                raise errors.AuthenticationError(
                    f"Please log in the Renku UI at {self._renku_url()} to complete authentication with Renku"
                )
            raise errors.AuthenticationError(
                "Please run the renku login command to authenticate with Renku or to refresh your expired credentials."
            )
        return response

    @staticmethod
    def _project_name_from_full_project_name(project_name: str) -> str:
        """Get just project name of project name if in owner/name form."""
        if "/" not in project_name:
            return project_name
        return project_name.rsplit("/", 1)[1]

    @property
    def name(self) -> str:
        """Return session provider's name."""
        return "renkulab"

    def is_remote_provider(self) -> bool:
        """Return True for remote providers (i.e. not local Docker)."""
        return True

    def build_image(self, image_descriptor: Path, image_name: str, config: Optional[Dict[str, Any]]):
        """Builds the container image."""
        if self.find_image(image_name, config=config):
            return
        repository = project_context.repository
        if repository.head.commit.hexsha != self._remote_head_hexsha():
            repository.push()
        self._wait_for_image(image_name=image_name, config=config)

    def find_image(self, image_name: str, config: Optional[Dict[str, Any]]) -> bool:
        """Find the given container image."""
        return (
            self._send_renku_request(
                "get",
                f"{self._notebooks_url()}/images",
                headers=self._auth_header(),
                params={"image_url": image_name},
            ).status_code
            == 200
        )

    @hookimpl
    def session_provider(self) -> IHibernatingSessionProvider:
        """Supported session provider.

        Returns:
            a reference to ``self``.
        """
        return self

    def get_start_parameters(self) -> List["ProviderParameter"]:
        """Returns parameters that can be set for session start."""
        from renku.core.dataset.providers.models import ProviderParameter

        return [
            ProviderParameter("ssh", help="Enable ssh connections to the session.", is_flag=True),
        ]

    def get_open_parameters(self) -> List["ProviderParameter"]:
        """Returns parameters that can be set for session open."""
        from renku.core.dataset.providers.models import ProviderParameter

        return [
            ProviderParameter("ssh", help="Open a remote terminal through SSH.", is_flag=True),
        ]

    def session_list(self, project_name: str, ssh_garbage_collection: bool = True) -> List[Session]:
        """Lists all the sessions currently running by the given session provider.

        Returns:
            list: a list of sessions.
        """
        sessions_res = self._send_renku_request(
            "get",
            f"{self._notebooks_url()}/servers",
            headers=self._auth_header(),
            params=self._get_renku_project_name_parts(),
        )
        if sessions_res.status_code == 200:
            sessions = [
                Session(
                    id=session["name"],
                    status=session.get("status", {}).get("state", "unknown"),
                    url=self.session_url(session["name"]),
                    start_time=datetime.fromisoformat(session.get("started")),
                    commit=session.get("annotations", {}).get("renku.io/commit-sha"),
                    branch=session.get("annotations", {}).get("renku.io/branch"),
                    provider="renkulab",
                    ssh_enabled=get_value("renku", "ssh_supported") == "true"
                    or project_context.project.template_metadata.ssh_supported,
                )
                for session in sessions_res.json().get("servers", {}).values()
            ]
            if ssh_garbage_collection:
                self._cleanup_ssh_connection_configs(project_name, running_sessions=sessions)
            return sessions
        return []

    def session_start(
        self,
        image_name: str,
        project_name: str,
        config: Optional[Dict[str, Any]],
        cpu_request: Optional[float] = None,
        mem_request: Optional[str] = None,
        disk_request: Optional[str] = None,
        gpu_request: Optional[str] = None,
        ssh: bool = False,
        **kwargs,
    ) -> Tuple[str, str]:
        """Creates an interactive session.

        Returns:
            Tuple[str, str]: Provider message and a possible warning message.
        """
        ssh_supported = (
            get_value("renku", "ssh_supported") == "true" or project_context.project.template_metadata.ssh_supported
        )
        if ssh and not ssh_supported:
            raise errors.RenkulabSessionError(
                "Cannot start session with SSH support because this project doesn't support SSH."
            )

        repository = project_context.repository

        session_commit = repository.head.commit.hexsha
        if repository.head.commit.hexsha != self._remote_head_hexsha():
            # INFO: The user is registered, the image is pinned or already available
            # but the local repository is not fully in sync with the remote
            communication.confirm(
                "You have unpushed commits that will not be present in your session. "
                "Renku can automatically push these commits so that they are present "
                "in the session you are launching. Do you wish to proceed?",
                abort=True,
            )
            repository.push()
        if ssh:
            self._cleanup_ssh_connection_configs(project_name)

        server_options: Dict[str, Union[str, float]] = {}
        if cpu_request:
            server_options["cpu_request"] = cpu_request
        if mem_request:
            server_options["mem_request"] = mem_request
        if gpu_request:
            server_options["gpu_request"] = int(gpu_request)
        if disk_request:
            server_options["disk_request"] = disk_request
        payload = {
            "image": image_name,
            "commit_sha": session_commit,
            "serverOptions": server_options,
            "branch": repository.active_branch.name if repository.active_branch else "master",
            **self._get_renku_project_name_parts(),
        }
        res = self._send_renku_request(
            "post",
            f"{self._notebooks_url()}/servers",
            headers=self._auth_header(),
            json=payload,
        )
        if res.status_code in [200, 201]:
            session_name = res.json()["name"]
            self._wait_for_session_status(session_name, "running")
            if ssh:
                name = self._project_name_from_full_project_name(project_name)
                connection = SystemSSHConfig().setup_session_config(name, session_name)
                return (
                    f"Session {session_name} successfully started, use 'renku session open --ssh {session_name}'"
                    f" or 'ssh {connection}' to connect to it",
                    "",
                )
            return (
                f"Session {session_name} successfully started, use 'renku session open {session_name}'"
                " to connect to it",
                "",
            )
        raise errors.RenkulabSessionError("Cannot start session via the notebook service because " + res.text)

    def session_stop(self, project_name: str, session_name: Optional[str], stop_all: bool) -> SessionStopStatus:
        """Stops all sessions (for the given project) or a specific interactive session."""
        responses = []
        sessions = self.session_list(project_name=project_name)
        n_sessions = len(sessions)

        if n_sessions == 0:
            return SessionStopStatus.NO_ACTIVE_SESSION

        if stop_all:
            for session in sessions:
                responses.append(
                    self._send_renku_request(
                        "delete", f"{self._notebooks_url()}/servers/{session.id}", headers=self._auth_header()
                    )
                )
                self._wait_for_session_status(session.id, "stopping")
        elif session_name:
            responses.append(
                self._send_renku_request(
                    "delete", f"{self._notebooks_url()}/servers/{session_name}", headers=self._auth_header()
                )
            )
            self._wait_for_session_status(session_name, "stopping")
        elif n_sessions == 1:
            responses.append(
                self._send_renku_request(
                    "delete", f"{self._notebooks_url()}/servers/{sessions[0].id}", headers=self._auth_header()
                )
            )
            self._wait_for_session_status(sessions[0].id, "stopping")
        else:
            return SessionStopStatus.NAME_NEEDED

        self._cleanup_ssh_connection_configs(project_name)

        n_successfully_stopped = len([r for r in responses if r.status_code == 204])

        return SessionStopStatus.SUCCESSFUL if n_successfully_stopped == n_sessions else SessionStopStatus.FAILED

    def session_open(self, project_name: str, session_name: Optional[str], ssh: bool = False, **kwargs) -> bool:
        """Open a given interactive session.

        Args:
            project_name(str): Renku project name.
            session_name(Optional[str]): The unique id of the interactive session.
            ssh(bool): Whether to open an SSH connection or a normal browser interface.
        """
        sessions = self.session_list(project_name="")
        system_config = SystemSSHConfig()
        name = self._project_name_from_full_project_name(project_name)
        ssh_prefix = f"{system_config.renku_host}-{name}-"

        if not session_name:
            if len(sessions) == 1:
                session_name = sessions[0].id
            else:
                return False

        if session_name.startswith(ssh_prefix):
            # NOTE: User passed in ssh connection name instead of session id by accident
            session_name = session_name.replace(ssh_prefix, "", 1)

        if not any(s.id == session_name for s in sessions):
            return False

        if ssh:
            ssh_setup = True
            if not system_config.is_configured:
                raise errors.RenkulabSessionError(
                    "SSH not set up for session. Run without '--ssh' or "
                    "run 'renku session ssh-setup' and start the session again."
                )
            elif not system_config.session_config_path(name, session_name).exists():
                # NOTE: Session wasn't launched from CLI
                if system_config.is_session_configured(session_name):
                    raise errors.RenkulabSessionError(
                        "Session wasn't started using 'renku session start --ssh ...' "
                        "and is not configured for SSH access by you."
                    )
                communication.info(f"Setting up SSH connection config for session {session_name}")
                system_config.setup_session_config(name, session_name)
                ssh_setup = False

            exit_code = pty.spawn(["ssh", session_name])

            if exit_code > 0 and not ssh_setup:
                # NOTE: We tried to connect to SSH even though it wasn't started from CLI
                # This failed, so we'll remove the temporary connection information.
                if system_config.session_config_path(project_name, session_name).exists():
                    system_config.session_config_path(project_name, session_name).unlink()
        else:
            url = self.session_url(session_name)

            webbrowser.open(url)
        return True

    def session_url(self, session_name: str) -> str:
        """Get the URL of the interactive session."""
        project_name_parts = self._get_renku_project_name_parts()
        session_url_parts = [
            "projects",
            project_name_parts["namespace"],
            project_name_parts["project"],
            "sessions/show",
            session_name,
        ]
        return urllib.parse.urljoin(self._renku_url(), "/".join(session_url_parts))

    def force_build_image(self, **kwargs) -> bool:
        """Whether we should force build the image directly or check for an existing image first."""
        return self._force_build

    def session_pause(self, project_name: str, session_name: Optional[str], **_) -> SessionStopStatus:
        """Pause all sessions (for the given project) or a specific interactive session."""

        def pause(session_name: str):
            result = self._send_renku_request(
                "patch",
                f"{self._notebooks_url()}/servers/{session_name}",
                headers=self._auth_header(),
                json={"state": "hibernated"},
            )

            self._wait_for_session_status(session_name, "hibernated")

            return result

        sessions = self.session_list(project_name=project_name)
        n_sessions = len(sessions)

        if n_sessions == 0:
            return SessionStopStatus.NO_ACTIVE_SESSION

        if session_name:
            response = pause(session_name)
        elif n_sessions == 1:
            response = pause(sessions[0].name)
        else:
            return SessionStopStatus.NAME_NEEDED

        return SessionStopStatus.SUCCESSFUL if response.status_code == 204 else SessionStopStatus.FAILED

    def session_resume(self, project_name: str, session_name: Optional[str], **kwargs) -> bool:
        """Resume a paused session.

        Args:
            project_name(str): Renku project name.
            session_name(Optional[str]): The unique id of the interactive session.
        """
        sessions = self.session_list(project_name="")
        system_config = SystemSSHConfig()
        name = self._project_name_from_full_project_name(project_name)
        ssh_prefix = f"{system_config.renku_host}-{name}-"

        if not session_name:
            if len(sessions) == 1:
                session_name = sessions[0].name
            else:
                return False
        else:
            if session_name.startswith(ssh_prefix):
                # NOTE: User passed in ssh connection name instead of session id by accident
                session_name = session_name.replace(ssh_prefix, "", 1)

            if not any(s.name == session_name for s in sessions):
                return False

        self._send_renku_request(
            "patch",
            f"{self._notebooks_url()}/servers/{session_name}",
            headers=self._auth_header(),
            json={"state": "running"},
        )

        self._wait_for_session_status(session_name, "running")

        return True



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Log of renku commands."""

from typing import List

from pydantic import ConfigDict, validate_call

from renku.command.command_builder import Command, inject
from renku.command.view_model.log import LogViewModel
from renku.core.interface.activity_gateway import IActivityGateway
from renku.core.interface.dataset_gateway import IDatasetGateway
from renku.domain_model.dataset import Dataset


def log_command():
    """Return a command for getting a log of renku commands."""
    return Command().command(_log)


@inject.autoparams("activity_gateway", "dataset_gateway")
@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def _log(
    activity_gateway: IActivityGateway,
    dataset_gateway: IDatasetGateway,
    workflows_only: bool = False,
    datasets_only: bool = False,
) -> List[LogViewModel]:
    """Get a log of renku commands.

    Args:
        activity_gateway(IActivityGateway): Injected activity gateway.
        dataset_gateway(IDatasetGateway): Injected dataset gateway.
        workflows_only(bool, optional): Whether to show logs for workflows only (Default value = False).
        datasets_only(bool, optional): Whether to show logs for datasets only (Default value = False).

    Returns:
        List of log entries.
    """

    def _get_all_dataset_versions(dataset: Dataset):
        """Get all datasets provenance for a dataset.

        Args:
            dataset(Dataset): Head dataset to get versions for.

        Returns:
            Iterator of all versions of a dataset.
        """

        current_dataset = dataset

        while current_dataset:
            yield current_dataset

            if current_dataset.is_derivation():
                assert current_dataset.derived_from is not None
                current_dataset = dataset_gateway.get_by_id(current_dataset.derived_from.value)
            else:
                return

    log_entries: List[LogViewModel] = []

    if not datasets_only:
        activities = activity_gateway.get_all_activities()
        log_entries.extend(LogViewModel.from_activity(a) for a in activities)

    if not workflows_only:
        datasets = [
            d
            for root_dataset in dataset_gateway.get_provenance_tails()
            for d in _get_all_dataset_versions(root_dataset)
        ]
        log_entries.extend(LogViewModel.from_dataset(d) for d in datasets)

    return log_entries



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku project/dataset image management."""

import os
import urllib
from pathlib import Path
from typing import List, Optional, Union, cast
from urllib.request import urlretrieve

import filetype

from renku.core import errors
from renku.core.constant import FILESYSTEM_ROOT
from renku.core.util.os import is_subpath
from renku.core.util.urls import is_remote
from renku.domain_model.image import ImageObject
from renku.domain_model.project_context import project_context


class ImageObjectRequest:
    """Model for passing image information."""

    def __init__(
        self,
        content_url: str,
        position: int = 0,
        mirror_locally: bool = True,
        safe_image_paths: Optional[List[str]] = None,
    ) -> None:
        self.content_url = content_url
        self.position = position
        self.mirror_locally = mirror_locally
        self.safe_image_paths: List[Union[str, Path]] = cast(List[Union[str, Path]], safe_image_paths) or []

    def to_image_object(self, owner_id: str) -> ImageObject:
        """Convert request model to ``ImageObject`` and download the image if requested and return its path."""
        self.safe_image_paths.append(project_context.path)

        if is_remote(self.content_url):
            if not self.mirror_locally:
                return ImageObject(
                    content_url=self.content_url,
                    position=self.position,
                    id=ImageObject.generate_id(owner_id=owner_id, position=self.position),
                )

            # NOTE: Download the image
            try:
                tmp_path, _ = urlretrieve(self.content_url)
            except urllib.error.URLError as e:
                raise errors.ImageError(f"Cannot download image with url {self.content_url}: {e}") from e

            path = Path(tmp_path)
        else:
            path = Path(self.content_url).resolve()

            if not os.path.exists(path):
                raise errors.ImageError(f"Image with local path '{self.content_url}' not found")
            # NOTE: Prevent path traversal or usage of non-image files
            elif (FILESYSTEM_ROOT in self.safe_image_paths and filetype.guess(path) is None) or not any(
                is_subpath(path, base=p) for p in self.safe_image_paths
            ):
                raise errors.ImageError(f"'{self.content_url}' isn't a valid image file")

        return ImageObject(
            content_url=path.as_posix(),
            position=self.position,
            id=ImageObject.generate_id(owner_id=owner_id, position=self.position),
        )



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""The base command for interacting with the Renku platform.

renku (base command)
********************

To list the available commands, either run ``renku`` with no parameters or
execute ``renku help``:

.. code-block:: console

    $ renku help
    Usage: renku [OPTIONS] COMMAND [ARGS]...

    Check common Renku commands used in various situations.


    Options:
      --version                       Print version number.
      --global-config-path            Print global application's config path.
      --path <path>                   Location of a Renku repository.
                                      [default: (dynamic)]
      --external-storage / -S, --no-external-storage
                                      Use an external file storage service.
      -h, --help                      Show this message and exit.

    Commands:
      # [...]

Configuration files
~~~~~~~~~~~~~~~~~~~

Depending on your system, you may find the configuration files used by Renku
command line in a different folder. By default, the following rules are used:

MacOS:
  ``~/Library/Application Support/Renku``
Unix:
  ``~/.config/renku``
Windows:
  ``C:\Users\<user>\AppData\Roaming\Renku``

If in doubt where to look for the configuration file, you can display its path
by running ``renku --global-config-path``.

.. cheatsheet::
   :group: Typical Workflow
   :command: $ git status
   :description: Take a look at what you have done since the last save.
   :target: ui,rp

.. cheatsheet::
   :group: Typical Workflow
   :command: $ renku save -m <msg>
   :description: Save your latest work, providing a message explaining what you have done.
   :target: ui,rp

.. cheatsheet::
   :group: Typical Workflow
   :command: $ renku run …
   :description: Run your code, capturing lineage of the inputs and outputs using Renku.
   :target: ui,rp

"""
import os
import sys
import uuid
from pathlib import Path

import click
import yaml
from click_plugins import with_plugins

from renku.command.options import option_external_storage_requested
from renku.command.util import WARNING
from renku.command.version import print_version
from renku.core import errors
from renku.core.constant import DATABASE_PATH
from renku.core.util.git import get_git_path
from renku.domain_model.project_context import project_context
from renku.ui.cli.clone import clone
from renku.ui.cli.config import config
from renku.ui.cli.dataset import dataset
from renku.ui.cli.doctor import doctor
from renku.ui.cli.env import env
from renku.ui.cli.exception_handler import IssueFromTraceback
from renku.ui.cli.gc import gc
from renku.ui.cli.githooks import githooks as githooks_command
from renku.ui.cli.graph import graph
from renku.ui.cli.init import init
from renku.ui.cli.log import log
from renku.ui.cli.login import credentials, login, logout
from renku.ui.cli.mergetool import mergetool
from renku.ui.cli.migrate import check_immutable_template_files, migrate, migrationscheck
from renku.ui.cli.move import move
from renku.ui.cli.project import project
from renku.ui.cli.remove import remove
from renku.ui.cli.rerun import rerun
from renku.ui.cli.rollback import rollback
from renku.ui.cli.run import run
from renku.ui.cli.save import save
from renku.ui.cli.service import service
from renku.ui.cli.session import session
from renku.ui.cli.status import status
from renku.ui.cli.storage import storage
from renku.ui.cli.template import template
from renku.ui.cli.update import update
from renku.ui.cli.workflow import workflow

try:
    from importlib.metadata import entry_points
except ImportError:
    from importlib_metadata import entry_points  # type: ignore


def get_entry_points(name: str):
    """Get entry points from importlib."""
    all_entry_points = entry_points()

    if hasattr(all_entry_points, "select"):
        return all_entry_points.select(group=name)  # type: ignore
    else:
        # Prior to Python 3.10, this returns a dict instead of the selection interface, which is slightly slower
        return all_entry_points.get(name, [])


WARNING_UNPROTECTED_COMMANDS = ["clone", "credentials", "env", "help", "init", "login", "logout", "service", "template"]

WARNING_UNPROTECTED_SUBCOMMANDS = {"template": ["ls", "show", "validate"]}


def _uuid_representer(dumper, data):
    """Add UUID serializer for YAML."""
    return dumper.represent_str(str(data))


def _is_renku_project(path: Path) -> bool:
    """Check if a path is a renku project."""
    from renku.core.constant import RENKU_HOME
    from renku.core.migration.utils import OLD_METADATA_PATH
    from renku.infrastructure.database import Database

    metadata_path = Path(path) / RENKU_HOME
    old_metadata = metadata_path / OLD_METADATA_PATH
    new_metadata = metadata_path / DATABASE_PATH / Database.ROOT_OID

    return old_metadata.exists() or new_metadata.exists()


yaml.add_representer(uuid.UUID, _uuid_representer)


def print_global_config_path(ctx, _, value):
    """Print global application's config path."""
    if not value or ctx.resilient_parsing:
        return

    click.echo(project_context.global_config_path)
    ctx.exit()


def is_allowed_subcommand(ctx):
    """Called from subcommands to check if their sub-subcommand is allowed.

    Subcommands where some sub-subcommands are allowed should be added to ``WARNING_UNPROTECTED_COMMANDS`` so they pass
    through the parent check and then added to ``WARNING_UNPROTECTED_SUBCOMMANDS`` so they get checked here.
    """
    from renku.domain_model.project_context import project_context

    if not _is_renku_project(project_context.path) and (
        not WARNING_UNPROTECTED_SUBCOMMANDS.get(ctx.command.name, False)
        or ctx.invoked_subcommand not in WARNING_UNPROTECTED_SUBCOMMANDS[ctx.command.name]
    ):
        raise errors.UsageError(
            f"{project_context.path} is not a renku repository.\n"
            "To initialize this as a renku repository use: 'renku init'"
        )


def is_allowed_command(ctx):
    """Check if invoked command contains help command."""

    return ctx.invoked_subcommand in WARNING_UNPROTECTED_COMMANDS or "-h" in sys.argv or "--help" in sys.argv


@with_plugins(get_entry_points("renku.cli_plugins"))
@click.group(
    cls=IssueFromTraceback, context_settings={"auto_envvar_prefix": "RENKU", "help_option_names": ["-h", "--help"]}
)
@click.option(
    "--version", is_flag=True, callback=print_version, expose_value=False, is_eager=True, help=print_version.__doc__
)
@click.option(
    "--global-config-path",
    is_flag=True,
    callback=print_global_config_path,
    expose_value=False,
    is_eager=True,
    help=print_global_config_path.__doc__,
)
@click.option(
    "--path", show_default=True, metavar="<path>", default=get_git_path, help="Location of a Renku repository."
)
@option_external_storage_requested
@click.pass_context
def cli(ctx, path, external_storage_requested):
    """Check common Renku commands used in various situations."""
    from renku.domain_model.project_context import project_context

    path = Path(path)

    is_command_allowed = is_allowed_command(ctx)
    is_renku_project = _is_renku_project(path)

    if not is_renku_project and not is_command_allowed:
        raise errors.UsageError(
            f"{path} is not a renku repository.\n" "To initialize this as a renku repository use: 'renku init'"
        )

    project_context.push_path(path)
    project_context.external_storage_requested = external_storage_requested

    if is_renku_project and path != Path(os.getcwd()) and not is_command_allowed:
        click.secho(WARNING + "Run CLI commands only from project's root directory.\n", err=True)


@cli.command()
@click.pass_context
def help(ctx):
    """Show help message and exit."""
    click.echo(ctx.parent.get_help())


# Register subcommands:
cli.add_command(check_immutable_template_files)
cli.add_command(clone)
cli.add_command(config)
cli.add_command(credentials)
cli.add_command(dataset)
cli.add_command(doctor)
cli.add_command(env)
cli.add_command(gc)
cli.add_command(githooks_command)
cli.add_command(graph)
cli.add_command(init)
cli.add_command(log)
cli.add_command(login)
cli.add_command(logout)
cli.add_command(mergetool)
cli.add_command(migrate)
cli.add_command(migrationscheck)
cli.add_command(move)
cli.add_command(project)
cli.add_command(remove)
cli.add_command(rerun)
cli.add_command(rollback)
cli.add_command(run)
cli.add_command(save)
cli.add_command(service)
cli.add_command(session)
cli.add_command(status)
cli.add_command(storage)
cli.add_command(template)
cli.add_command(update)
cli.add_command(workflow)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Project class."""

from dataclasses import dataclass
from datetime import datetime
from typing import TYPE_CHECKING, Dict, List, Optional, cast
from urllib.parse import quote

import persistent

from renku.core import errors
from renku.core.util.datetime8601 import fix_datetime, local_now, parse_date
from renku.core.util.git import get_git_user
from renku.core.util.os import normalize_to_ascii
from renku.domain_model.constant import NO_VALUE
from renku.domain_model.image import ImageObject
from renku.domain_model.provenance.agent import Person
from renku.domain_model.provenance.annotation import Annotation
from renku.version import __minimum_project_version__

if TYPE_CHECKING:
    from renku.domain_model.project_context import ProjectContext, ProjectRemote
    from renku.infrastructure.repository import Repository


@dataclass
class ProjectTemplateMetadata:
    """Metadata about the template used in a project."""

    template_id: Optional[str] = None
    metadata: str = ""
    template_ref: Optional[str] = None
    template_source: Optional[str] = None
    template_version: Optional[str] = None
    immutable_template_files: Optional[List[str]] = None
    ssh_supported: bool = False


class Project(persistent.Persistent):
    """Represent a project."""

    keywords: List[str] = list()
    image: Optional[ImageObject] = None

    # NOTE: the minimum version of renku to needed to work with a project
    # This should be bumped on metadata version changes and when we do not forward-compatible on-the-fly migrations
    minimum_renku_version: str = __minimum_project_version__

    def __init__(
        self,
        *,
        agent_version: Optional[str] = None,
        annotations: Optional[List[Annotation]] = None,
        creator: Person,
        date_created: Optional[datetime] = None,
        description: Optional[str] = None,
        id: Optional[str] = None,
        name: Optional[str] = None,
        template_metadata: Optional[ProjectTemplateMetadata] = None,
        version: Optional[str] = None,
        keywords: Optional[List[str]] = None,
        image: Optional[ImageObject] = None,
    ):
        from renku.core.migration.migrate import SUPPORTED_PROJECT_VERSION

        version = cast(str, version or SUPPORTED_PROJECT_VERSION)
        date_created = parse_date(date_created) or local_now()

        if id is None:
            namespace, generated_name = Project.get_namespace_and_name(name=name, creator=creator)
            assert generated_name is not None, "Cannot generate Project id with no name"
            id = Project.generate_id(namespace=namespace, name=generated_name)

        self.name: Optional[str] = name
        self.agent_version: Optional[str] = agent_version
        self.annotations: List[Annotation] = annotations or []
        self.creator: Person = creator
        self.date_created: datetime = fix_datetime(date_created) or local_now()
        self.description: Optional[str] = description
        self.id: str = id
        self.version: str = version
        self.keywords = keywords or []
        self.image = image

        self.template_metadata: ProjectTemplateMetadata = template_metadata or ProjectTemplateMetadata()

        # NOTE: We copy this over as class variables don't get saved in the DB
        self.minimum_renku_version = Project.minimum_renku_version

    @classmethod
    def from_project_context(
        cls,
        project_context: "ProjectContext",
        name: Optional[str] = None,
        namespace: Optional[str] = None,
        description: Optional[str] = None,
        keywords: Optional[List[str]] = None,
        custom_metadata: Optional[Dict] = None,
        creator: Optional[Person] = None,
        image: Optional[ImageObject] = None,
    ) -> "Project":
        """Create an instance from a path.

        Args:
            cls: The class.
            name(Optional[str]): Name of the project (when creating a new one) (Default value = None).
            namespace(Optional[str]): Namespace of the project (when creating a new one) (Default value = None).
            description(Optional[str]): Project description (when creating a new one) (Default value = None).
            keywords(Optional[List[str]]): Keywords for the project (when creating a new one) (Default value = None).
            custom_metadata(Optional[Dict]): Custom JSON-LD metadata (when creating a new project)
                (Default value = None).
            creator(Optional[Person]): The project creator.
            image(Optional[ImageObject]): Project's image/avatar.
        """
        creator = creator or get_git_user(repository=project_context.repository)

        namespace, name = cls.get_namespace_and_name(
            remote=project_context.remote, name=name, namespace=namespace, creator=creator
        )
        annotations = None

        if custom_metadata:
            annotations = [Annotation(id=Annotation.generate_id(), body=custom_metadata, source="renku")]

        if creator is None:
            raise errors.ParameterError("Project Creator not set", "creator")

        if name is None:
            raise errors.ParameterError("Project 'name' not set and could not be generated", "name")

        if namespace is None:
            raise errors.ParameterError("Project 'namespace' not set and could not be generated", "namespace")

        id = cls.generate_id(namespace=namespace, name=name)
        return cls(
            annotations=annotations,
            creator=creator,
            description=description,
            id=id,
            image=image,
            keywords=keywords,
            name=name,
        )

    @staticmethod
    def get_namespace_and_name(
        *,
        remote: Optional["ProjectRemote"] = None,
        repository: Optional["Repository"] = None,
        name: Optional[str] = None,
        namespace: Optional[str] = None,
        creator: Optional[Person] = None,
    ):
        """Return Project's namespace and name from various objects."""
        if remote:
            namespace = namespace or remote.owner
            name = name or remote.name

        if not creator and repository:
            creator = get_git_user(repository=repository)

        if not namespace and creator:
            namespace = creator.email.split("@")[0]

        return namespace, name

    @staticmethod
    def generate_id(namespace: str, name: str):
        """Generate an id for Project."""
        assert namespace, "Cannot generate Project id with no namespace"
        assert name, "Cannot generate Project id with no name"

        namespace = quote(namespace.strip("/"), safe="/")
        slug = normalize_to_ascii(name)

        return f"/projects/{namespace}/{slug}"

    def update_metadata(self, custom_metadata=None, custom_metadata_source=None, **kwargs):
        """Updates metadata."""
        editable_attributes = ["creator", "description", "keywords"]
        for name, value in kwargs.items():
            if name not in editable_attributes:
                raise errors.ParameterError(f"Cannot edit field: '{name}'")
            if value is not NO_VALUE and value != getattr(self, name):
                setattr(self, name, value)

        if custom_metadata is not NO_VALUE and custom_metadata_source is not NO_VALUE:
            existing_metadata = [a for a in self.annotations if a.source != custom_metadata_source]
            if custom_metadata is not None:
                if not isinstance(custom_metadata, list):
                    custom_metadata = [custom_metadata]
                for icustom_metadata in custom_metadata:
                    existing_metadata.append(
                        Annotation(
                            id=Annotation.generate_id(),
                            body=icustom_metadata,
                            source=custom_metadata_source,
                        )
                    )

            self.annotations = existing_metadata



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Manage names of Renku objects."""

import os
import subprocess
from pathlib import Path

import attr

from renku.core import errors
from renku.core.constant import REFS
from renku.domain_model.project_context import project_context


@attr.s(slots=True)
class LinkReference:
    """Manage linked object names."""

    metadata_path = attr.ib(default=None, kw_only=True)
    name = attr.ib()

    @classmethod
    def check_ref_format(cls, name, no_slashes=False):
        r"""Ensures that a reference name is well formed.

        It follows Git naming convention:

        - any path component of it begins with ".", or
        - it has double dots "..", or
        - it has ASCII control characters, or
        - it has ":", "?", "\[", "\\", "^", "~", SP, or TAB anywhere, or
        - it has "*" anywhere, or
        - it ends with a "/", or
        - it ends with ".lock", or
        - it contains a "@{" portion
        """
        params = ("--allow-onelevel", name) if no_slashes else (name,)
        return subprocess.run(("git", "check-ref-format") + params).returncode == 0

    @name.validator
    def name_validator(self, attribute, value):
        """Validate reference name."""
        if not self.check_ref_format(value):
            raise errors.ParameterError(f'The reference name "{value}" is not valid.')

    @property
    def path(self):
        """Return full reference path."""
        return self.metadata_path / REFS / self.name

    @property
    def reference(self):
        """Return the path we point to."""
        return self.path.resolve()

    def delete(self):
        """Delete the reference at the given path."""
        return self.path.unlink()

    def set_reference(self, reference):
        """Set ourselves to the given reference path."""
        reference_path = Path(reference).resolve().absolute()
        reference_path.relative_to(project_context.path)
        self.path.parent.mkdir(parents=True, exist_ok=True)
        os.symlink(os.path.relpath(str(reference_path), start=str(self.path.parent)), str(self.path))

    @classmethod
    def iter_items(cls, common_path=None):
        """Find all references in the repository."""
        refs_path = path = project_context.metadata_path / REFS
        if common_path:
            path = path / common_path

        for name in path.rglob("*"):
            if name.is_dir():
                continue
            yield cls(metadata_path=project_context.metadata_path, name=str(name.relative_to(refs_path)))

    @classmethod
    def create(cls, name, force=False):
        """Create symlink to object in reference path."""
        ref = cls(metadata_path=project_context.metadata_path, name=name)
        path = ref.path
        if os.path.lexists(path):
            if not force:
                raise OSError(str(path))
            else:
                ref.delete()

        return ref

    def rename(self, new_name, force=False):
        """Rename self to a new name."""
        new_ref = self.create(name=new_name, force=force)
        new_ref.set_reference(self.reference)
        self.delete()
        return new_ref



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""DatasetFile source and url migrations."""

from renku.core.migration.models.v7 import get_project_datasets
from renku.core.migration.utils import generate_dataset_file_url


def migrate(_):
    """Migration function."""
    _fix_dataset_file_source_and_url()


def _fix_dataset_file_source_and_url():
    for dataset in get_project_datasets():
        for file in dataset.files:
            file.source = file.url
            file.url = generate_dataset_file_url(filepath=file.path)

            if file.source:
                file.source = file.source.replace("file://", "")

            if file.based_on:
                file.based_on.source = file.based_on.url

        dataset.to_yaml()



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""cwltool based provider.

This implementation provides a plugin for executing workflows using ``cwltool``,
by simply exporting the workflow into a CWL format and then using ``cwltool``
to run the exported CWL file.

.. code-block:: console

   $ renku workflow execute --provider cwltool example_workflow


.. topic:: Specifying custom parameters for cwltool (``--config``)

   You can specify custom configuration parameters for ``cwltool`` executer
   by providing a YAML file for the ``--config`` option. The YAML file
   should contain both the ``LoadingContext`` and ``RuntimeContext`` key-value
   pairs.

.. code-block:: console

   $ renku workflow execute --config config.yaml --provider cwltool example_workflow
"""

import os
import shutil
import sys
import tempfile
from pathlib import Path
from typing import Any, Dict
from urllib.parse import unquote
from uuid import uuid4

import cwltool.factory
import networkx as nx
from cwltool.context import LoadingContext, RuntimeContext

from renku.command.util import progressbar
from renku.core.errors import NodeNotFoundError, WorkflowExecuteError
from renku.core.plugin import hookimpl
from renku.core.plugin.workflow import workflow_converter
from renku.domain_model.workflow.composite_plan import CompositePlan
from renku.domain_model.workflow.provider import IWorkflowProvider


class CWLToolProvider(IWorkflowProvider):
    """A workflow executor provider using cwltool."""

    @hookimpl
    def workflow_provider(self):
        """Workflow provider name."""
        return self, "cwltool"

    @hookimpl
    def workflow_execute(self, dag: nx.DiGraph, basedir: Path, config: Dict[str, Any]):
        """Executes a given workflow using cwltool."""

        if not shutil.which("nodejs") and not shutil.which("node"):
            raise NodeNotFoundError()

        # run cwl with cwltool
        argv = sys.argv
        sys.argv = ["cwltool"]
        runtime_args = {"rm_tmpdir": False, "move_outputs": "leave", "preserve_entire_environment": True}
        loading_args = {"relax_path_checks": True}
        if config:
            # update both RuntimeContext and LoadingContext parameters with user supplied values
            # context.ContextBase takes care that only available parameters are set in a given class
            runtime_args.update(config)
            loading_args.update(config)

        # Keep all environment variables.
        runtime_context = RuntimeContext(kwargs=runtime_args)
        loading_context = LoadingContext(kwargs=loading_args)

        factory = cwltool.factory.Factory(loading_context=loading_context, runtime_context=runtime_context)

        output_paths = []

        workflows = list(dag.nodes)
        if len(workflows) > 1:
            workflow = CompositePlan(
                id=CompositePlan.generate_id(), plans=workflows, name=f"plan-collection-{uuid4().hex}"
            )
        else:
            workflow = workflows[0]

        with tempfile.NamedTemporaryFile() as f:
            # export Plan to cwl
            converter = workflow_converter("cwl")
            converter(
                workflow=workflow,
                basedir=basedir,
                output=Path(f.name),
                output_format=None,
                resolve_paths=True,
                nest_workflows=False,
            )

            process = factory.make(os.path.relpath(str(f.name)))

            try:
                outputs = process()
            except cwltool.factory.WorkflowStatus as e:
                raise WorkflowExecuteError() from e

            sys.argv = argv

            # Move outputs to correct location in the repository.
            output_dirs = process.factory.executor.output_dirs

            def remove_prefix(location, prefix="file://"):
                if location.startswith(prefix):
                    return unquote(location[len(prefix) :])
                return unquote(location)

            locations = {remove_prefix(output["location"]) for output in outputs.values()}  # type: ignore
            # make sure to not move an output if it's containing directory gets moved
            locations = {
                location
                for location in locations
                if not any(location.startswith(d) for d in locations if location != d)
            }

            with progressbar(locations, label="Moving outputs") as bar:
                for location in bar:
                    for output_dir in output_dirs:
                        if location.startswith(output_dir):
                            output_path = location[len(output_dir) :].lstrip(os.path.sep)
                            destination = basedir / output_path
                            output_paths.append(destination)
                            if destination.is_dir():
                                shutil.rmtree(str(destination))
                                destination = destination.parent
                                destination.mkdir(parents=True, exist_ok=True)
                            else:
                                destination.parent.mkdir(parents=True, exist_ok=True)
                            shutil.move(location, str(destination))
                            continue

        return output_paths



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interactive session engine."""

from __future__ import annotations

from abc import ABCMeta, abstractmethod
from datetime import datetime
from enum import Enum, auto
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple

from renku.core.constant import ProviderPriority

if TYPE_CHECKING:
    from renku.core.dataset.providers.models import ProviderParameter


class SessionStopStatus(Enum):
    """Status code returned when stopping/pausing sessions."""

    NO_ACTIVE_SESSION = auto()
    SUCCESSFUL = auto()
    FAILED = auto()  # When all or some of (requested) sessions can't be stopped
    NAME_NEEDED = auto()


class Session:
    """Interactive session."""

    def __init__(
        self,
        id: str,
        status: str,
        url: str,
        commit: str,
        branch: str,
        provider: str,
        start_time: datetime,
        ssh_enabled: bool = False,
    ):
        self.id = id
        self.status = status
        self.url = url
        self.start_time = start_time
        self.commit = commit
        self.branch = branch
        self.provider = provider
        self.ssh_enabled = ssh_enabled

    @property
    def name(self) -> str:
        """Return session name which is the same as its id."""
        return self.id


class ISessionProvider(metaclass=ABCMeta):
    """Abstract class for an interactive session provider."""

    priority: ProviderPriority = ProviderPriority.NORMAL

    @property
    @abstractmethod
    def name(self) -> str:
        """Return session provider's name."""
        pass

    @abstractmethod
    def is_remote_provider(self) -> bool:
        """Return True for remote providers (i.e. not local Docker)."""
        raise NotImplementedError

    @abstractmethod
    def build_image(self, image_descriptor: Path, image_name: str, config: Optional[Dict[str, Any]]):
        """Builds the container image.

        Args:
            image_descriptor: Path to the container image descriptor file.
            image_name: Container image name.
            config: Path to the session provider specific configuration YAML.
        """
        pass

    @abstractmethod
    def find_image(self, image_name: str, config: Optional[Dict[str, Any]]) -> bool:
        """Search for the given container image.

        Args:
            image_name(str): Container image name.
            config: Path to the session provider specific configuration YAML.

        Returns:
            bool: True if the given container images is available locally.
        """
        pass

    @abstractmethod
    def session_provider(self) -> ISessionProvider:
        """Supported session provider.

        Returns:
            a reference to ``self``.
        """
        pass

    @abstractmethod
    def get_start_parameters(self) -> List[ProviderParameter]:
        """Returns parameters that can be set for session start."""
        pass

    @abstractmethod
    def get_open_parameters(self) -> List[ProviderParameter]:
        """Returns parameters that can be set for session open."""
        pass

    @abstractmethod
    def session_list(self, project_name: str) -> List[Session]:
        """Lists all the sessions currently running by the given session provider.

        Args:
            project_name(str): Renku project name.

        Returns:
            a list of sessions.
        """
        pass

    @abstractmethod
    def session_start(
        self,
        image_name: str,
        project_name: str,
        config: Optional[Dict[str, Any]],
        cpu_request: Optional[float] = None,
        mem_request: Optional[str] = None,
        disk_request: Optional[str] = None,
        gpu_request: Optional[str] = None,
        **kwargs,
    ) -> Tuple[str, str]:
        """Creates an interactive session.

        Args:
            image_name(str): Container image name to be used for the interactive session.
            project_name(str): The project identifier.
            config(Optional[Dict[str, Any]]): Path to the session provider specific configuration YAML.
            cpu_request(Optional[float]): CPU request for the session.
            mem_request(Optional[str]): Memory size request for the session.
            disk_request(Optional[str]): Disk size request for the session.
            gpu_request(Optional[str]): GPU device request for the session.

        Returns:
            Tuple[str, str]: Provider message and a possible warning message.
        """
        pass

    @abstractmethod
    def session_stop(self, project_name: str, session_name: Optional[str], stop_all: bool) -> SessionStopStatus:
        """Stops all or a given interactive session.

        Args:
            project_name(str): Project's name.
            session_name(str, optional): The unique id of the interactive session.
            stop_all(bool): Specifies whether to stop all the running interactive sessions.


        Returns:
            SessionStopStatus: The status of running and stopped sessions
        """
        pass

    @abstractmethod
    def session_open(self, project_name: str, session_name: Optional[str], **kwargs) -> bool:
        """Open a given interactive session.

        Args:
            project_name(str): Renku project name.
            session_name(Optional[str]): The unique id of the interactive session.
        """
        pass

    @abstractmethod
    def session_url(self, session_name: str) -> Optional[str]:
        """Get the given session's URL.

        Args:
            session_name(str): The unique id of the interactive session.

        Returns:
            URL of the interactive session.
        """
        pass

    def pre_start_checks(self, **kwargs):
        """Perform any required checks on the state of the repository prior to starting a session.

        The expectation is that this method will abort the
        session start if the checks are not successful or will take corrective actions to
        make sure that the session launches successfully. By default, this method does not do any checks.
        """
        return None

    def force_build_image(self, **kwargs) -> bool:
        """Whether we should force build the image directly or check for an existing image first."""
        return False


class IHibernatingSessionProvider(ISessionProvider):
    """Abstract class for an interactive session provider that supports hibernation."""

    @abstractmethod
    def session_pause(self, project_name: str, session_name: Optional[str], **kwargs) -> SessionStopStatus:
        """Pause all or a given interactive session.

        Args:
            project_name(str): Project's name.
            session_name(str, optional): The unique id of the interactive session.

        Returns:
            SessionStopStatus: The status of running and paused sessions
        """

    @abstractmethod
    def session_resume(self, project_name: str, session_name: Optional[str], **kwargs) -> bool:
        """Resume a paused session.

        Args:
            project_name(str): Renku project name.
            session_name(Optional[str]): The unique id of the interactive session.
        """



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Communicator class for printing click output."""

import sys
from contextlib import contextmanager

import click
from tqdm import tqdm
from yaspin import yaspin

import renku.ui.cli.utils.color as color
from renku.core.util.communication import CommunicationCallback


class StandardOutput(CommunicationCallback):
    """Communication listener that outputs to stdout/stderr."""

    def __init__(self):
        super().__init__()
        self._progress_bars = {}
        self._progress_types = ["download"]
        self._spinner = None

    @contextmanager
    def pause_spinner(self):
        """Pause Spinner so output gets actually written."""
        if self._spinner:
            self._spinner.hide()

        yield

        if self._spinner:
            self._spinner.show()

    def echo(self, msg, end="\n"):
        """Write a message."""
        with CommunicationCallback.lock, self.pause_spinner():
            print(msg, end=end)

    def info(self, msg):
        """Write an info message."""
        with CommunicationCallback.lock, self.pause_spinner():
            print(msg)

    def warn(self, msg):
        """Write a warning message."""
        with CommunicationCallback.lock, self.pause_spinner():
            print(msg)

    def error(self, msg):
        """Write an error message."""
        with CommunicationCallback.lock, self.pause_spinner():
            print(msg, file=sys.stderr)

    def confirm(self, msg, abort=False, warning=False, default=False):
        """Get confirmation for an action."""
        return False

    def start_progress(self, name, total, **kwargs):
        """Start a new tqdm progressbar."""
        if name in self._progress_bars:
            raise ValueError(f"Name {name} is already a registered progressbar.")

        if "type" not in kwargs:
            kwargs["type"] = "download"

        if kwargs["type"] not in self._progress_types:
            self._progress_bars[name] = None
        elif kwargs["type"] == "download":
            self._progress_bars[name] = tqdm(
                total=total,
                unit="iB",
                unit_scale=True,
                desc=name,
                leave=False,
                bar_format="{desc:.32}: {percentage:3.0f}%|{bar}{r_bar}",
            )

    def update_progress(self, name, amount):
        """Update a progressbar."""
        if name not in self._progress_bars or not self._progress_bars[name]:
            return

        self._progress_bars[name].update(amount)

    def finalize_progress(self, name):
        """End a progressbar."""
        if name not in self._progress_bars or not self._progress_bars[name]:
            return

        self._progress_bars[name].close()
        del self._progress_bars[name]

    @contextmanager
    def busy(self, msg):
        """Indicate busy status using a spinner."""
        self._spinner = yaspin(text=msg)
        try:
            self._spinner.start()
            yield
        finally:
            self._spinner.stop()
            self._spinner = None


class ClickCallback(StandardOutput):
    """CommunicationCallback implementation for ``click`` messages."""

    INFO = click.style("Info: ", bold=True, fg=color.BLUE)
    WARNING = click.style("Warning: ", bold=True, fg=color.YELLOW)
    ERROR = click.style("Error: ", bold=True, fg=color.RED)

    def echo(self, msg, end="\n"):
        """Write a message."""
        new_line = True
        if end != "\n":
            msg = msg + end
            new_line = False
        with self.pause_spinner():
            click.echo(msg, nl=new_line)

    def info(self, msg):
        """Write an info message."""
        with self.pause_spinner():
            click.echo(self.INFO + msg)

    def warn(self, msg):
        """Write a warning message."""
        with self.pause_spinner():
            click.echo(self.WARNING + msg)

    def error(self, msg):
        """Write an error message."""
        with self.pause_spinner():
            click.echo(self.ERROR + msg, err=True)

    def has_prompt(self):
        """Return True if communicator provides a direct prompt to users."""
        return True

    def confirm(self, msg, abort=False, warning=False, default=False):
        """Get confirmation for an action using a prompt."""
        prefix = self.WARNING if warning else ""
        with self.pause_spinner():
            return click.confirm(prefix + msg, abort=abort, default=default)

    def prompt(self, msg, type=None, default=None, **kwargs):
        """Show a message prompt from the first callback that has a prompt."""
        with self.pause_spinner():
            return click.prompt(msg, type=type, default=default, **kwargs)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Zenodo API integration."""

import json
import os
import posixpath
import urllib
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional
from urllib.parse import urlparse

from renku.core import errors
from renku.core.dataset.providers.api import (
    ExporterApi,
    ExportProviderInterface,
    ImportProviderInterface,
    ProviderApi,
    ProviderPriority,
)
from renku.core.dataset.providers.repository import RepositoryImporter, make_request
from renku.core.util import communication
from renku.core.util.doi import is_doi
from renku.core.util.requests import get_redirect_url
from renku.core.util.urls import remove_credentials
from renku.domain_model.project_context import project_context

if TYPE_CHECKING:
    from renku.core.dataset.providers.models import ProviderDataset, ProviderParameter
    from renku.domain_model.dataset import Dataset, DatasetTag


ZENODO_BASE_URL = "https://zenodo.org"
ZENODO_SANDBOX_URL = "https://sandbox.zenodo.org/"

ZENODO_API_PATH = "api"

ZENODO_DEPOSIT_PATH = "deposit"
ZENODO_PUBLISH_PATH = "record"

ZENODO_PUBLISH_ACTION_PATH = "depositions/{0}/actions/publish"
ZENODO_METADATA_URL = "depositions/{0}"
ZENODO_FILES_URL = "depositions/{0}/files"
ZENODO_NEW_DEPOSIT_URL = "depositions"


class ZenodoProvider(ProviderApi, ExportProviderInterface, ImportProviderInterface):
    """Zenodo registry API provider."""

    priority = ProviderPriority.HIGH
    name = "Zenodo"
    is_remote = True

    def __init__(self, uri: str, is_doi: bool = False):
        super().__init__(uri=uri)

        self.is_doi = is_doi
        self._publish: bool = False

    @staticmethod
    def supports(uri):
        """Whether this provider supports a given URI."""
        if "zenodo" in uri.lower():
            return True

        return False

    @staticmethod
    def get_record_id(uri):
        """Extract record id from URI."""
        parts = urlparse(uri).path.split("/")
        parts = [p for p in parts if p.isdigit()]
        return parts[-1]

    @staticmethod
    def get_export_parameters() -> List["ProviderParameter"]:
        """Returns parameters that can be set for export."""
        from renku.core.dataset.providers.models import ProviderParameter

        return [ProviderParameter("publish", help="Publish the exported dataset.", is_flag=True)]

    def get_importer(self, **kwargs) -> "ZenodoImporter":
        """Get importer for a record from Zenodo."""
        from renku.core.dataset.providers.doi import DOIProvider

        uri = self.uri
        if self.is_doi:
            # NOTE: Resolve the DOI and make a record for the retrieved record id
            doi = DOIProvider(uri=uri).get_importer()
            uri = doi.uri

        response = _make_request(uri)
        return ZenodoImporter(uri=uri, original_uri=self.uri, json=response.json())

    def get_exporter(
        self, dataset: "Dataset", *, tag: Optional["DatasetTag"], publish: bool = False, **kwargs
    ) -> "ZenodoExporter":
        """Create export manager for given dataset."""
        self._publish = publish
        return ZenodoExporter(dataset=dataset, publish=self._publish, tag=tag)


class ZenodoImporter(RepositoryImporter):
    """Zenodo importer."""

    def __init__(self, *, uri: str, original_uri, json: Dict[str, Any]):
        super().__init__(uri=uri, original_uri=original_uri)

        self._jsonld: Optional[dict] = None
        self._json = json

        metadata = self._json.pop("metadata", {})
        self._json["metadata"] = ZenodoMetadataSerializer.from_metadata(metadata) if metadata is not None else None
        record_id = self._json.pop("record_id", None) or self._json.pop("recid", None)
        self._json["record_id"] = str(record_id) if record_id is not None else None

        # NOTE: Make sure that these properties have a default value
        self._json["links"] = self._json.pop("links", {})
        self._json["files"] = self._json.pop("files", [])

    @property
    def version(self):
        """Get record version."""
        return self._json["metadata"].version

    @property
    def latest_uri(self):
        """Get URI of latest version."""
        return get_redirect_url(self._json["links"].get("latest"))

    def is_latest_version(self):
        """Check if this record is the latest version."""
        return ZenodoProvider.get_record_id(self.latest_uri) == self._json["record_id"]

    def get_jsonld(self):
        """Get record metadata as jsonld."""
        response = _make_request(self._uri, accept="application/ld+json")
        self._jsonld = response.json()

        if self._jsonld is not None and "image" in self._jsonld and isinstance(self._jsonld["image"], str):
            self._jsonld["image"] = {
                "@id": self._jsonld["image"],
                "@type": "ImageObject",
                "position": 1,
                "contentUrl": self._jsonld["image"],
            }

        return self._jsonld

    def get_files(self):
        """Get Zenodo files metadata as ``ZenodoFile``."""
        if not self._json["files"]:
            raise LookupError("no files have been found - deposit is empty or protected")

        return [ZenodoFileSerializer(**file) for file in self._json["files"]]

    def fetch_provider_dataset(self) -> "ProviderDataset":
        """Deserialize a `Dataset`."""
        from marshmallow import pre_load

        from renku.command.schema.agent import PersonSchema
        from renku.core.dataset.providers.models import ProviderDataset, ProviderDatasetFile, ProviderDatasetSchema
        from renku.domain_model.dataset import Url, generate_default_slug

        class ZenodoDatasetSchema(ProviderDatasetSchema):
            """Schema for Zenodo datasets."""

            @pre_load
            def fix_data(self, data, **_):
                """Fix data that is received from Zenodo."""
                # Fix context
                context = data.get("@context")
                if context and isinstance(context, str):
                    if not context.endswith("/"):
                        context = f"{context}/"
                    if context == "https://schema.org/":
                        context = "http://schema.org/"
                    data["@context"] = {"@base": context, "@vocab": context}
                # Add type to creators
                creators = data.get("creator", [])
                for c in creators:
                    c["@type"] = [str(t) for t in PersonSchema.opts.rdf_type]

                # Fix license to be a string
                license = data.get("license")
                if license and isinstance(license, dict):
                    data["license"] = license.get("url", "")
                # fix keywords to be a list
                keywords = data.get("keywords")
                if keywords and isinstance(keywords, str):
                    data["keywords"] = [k.strip() for k in keywords.split(",")]

                # Delete existing isPartOf
                data.pop("isPartOf", None)
                data.pop("sameAs", None)

                return data

        files = self.get_files()
        metadata = self.get_jsonld()
        dataset = ProviderDataset.from_jsonld(metadata, schema_class=ZenodoDatasetSchema)
        dataset.slug = generate_default_slug(name=dataset.name or "", version=dataset.version)
        dataset.same_as = Url(url_id=remove_credentials(self.original_uri))
        if is_doi(dataset.identifier):
            dataset.same_as = Url(url_str=urllib.parse.urljoin("https://doi.org", dataset.identifier))

        self._provider_dataset_files = [
            ProviderDatasetFile(
                source=file.remote_url.geturl(),
                filename=Path(file.filename).name,
                checksum=file.checksum,
                filesize=file.filesize,
                filetype=file.type,
                path="",
            )
            for file in files
        ]

        self._provider_dataset = dataset
        return self._provider_dataset


class ZenodoFileSerializer:
    """Zenodo record file."""

    def __init__(self, *, id=None, checksum=None, links=None, key=None, size=None, **_):
        self.id = id
        self.checksum = checksum
        self.links = links
        self.filename = key
        self.filesize = size

    @property
    def remote_url(self):
        """Get remote URL as ``urllib.ParseResult``."""
        return urllib.parse.urlparse(self.links["self"])

    @property
    def type(self):
        """Get file type."""
        return self.filename.split(".")[-1]


class ZenodoMetadataSerializer:
    """Zenodo metadata."""

    def __init__(
        self,
        *,
        access_right=None,
        communities=None,
        contributors=None,
        creators=None,
        description=None,
        doi=None,
        extras=None,
        grants=None,
        image_type=None,
        journal_issue=None,
        journal_pages=None,
        journal_title=None,
        journal_volume=None,
        keywords=None,
        language=None,
        license=None,
        notes=None,
        prereserve_doi=None,
        publication_date=None,
        publication_type=None,
        references=None,
        related_identifiers=None,
        title=None,
        upload_type=None,
        version=None,
    ):
        self.access_right = access_right
        self.communities = communities
        self.contributors = contributors
        self.creators = creators
        self.description = description
        self.doi = doi
        self.extras = extras
        self.grants = grants
        self.image_type = image_type
        self.journal_issue = journal_issue
        self.journal_pages = journal_pages
        self.journal_title = journal_title
        self.journal_volume = journal_volume
        self.keywords = keywords
        self.language = language
        self.license = license
        self.notes = notes
        self.prereserve_doi = prereserve_doi
        self.publication_date = publication_date
        self.publication_type = publication_type
        self.references = references
        self.related_identifiers = related_identifiers
        self.title = title
        self.upload_type = upload_type
        self.version = version

    @classmethod
    def from_metadata(cls, metadata: Dict[str, Any]) -> "ZenodoMetadataSerializer":
        """Create an instance from a metadata dict.

        Args:
            metadata: The dict data to convert.

        Returns:
            ZenodoMetadataSerializer: Serializer containing data in deserialized form.
        """
        all_keys = set(vars(ZenodoMetadataSerializer()).keys())

        data = {key: metadata.get(key) for key in all_keys}
        data["extras"] = {key: metadata.get(key) for key in (metadata.keys() - all_keys)}

        return ZenodoMetadataSerializer(**data)


class ZenodoExporter(ExporterApi):
    """Zenodo export manager."""

    HEADERS = {
        "Content-Type": "application/json",
        "Referer": f"https://{os.environ.get('RENKU_DOMAIN', 'zenodo.org')}",
    }

    def __init__(self, dataset, publish, tag):
        super().__init__(dataset)
        self._access_token = None
        self._publish = publish
        self._tag = tag

    @property
    def zenodo_url(self):
        """Returns correct Zenodo URL based on environment."""
        return ZENODO_SANDBOX_URL if "ZENODO_USE_SANDBOX" in os.environ else ZENODO_BASE_URL

    def set_access_token(self, access_token):
        """Set access token."""
        self._access_token = access_token

    def get_access_token_url(self):
        """Endpoint for creation of access token."""
        return urllib.parse.urlparse("https://zenodo.org/account/settings/applications/tokens/new/").geturl()

    @property
    def default_params(self):
        """Create request default parameters."""
        return {"access_token": self._access_token}

    def dataset_to_request(self):
        """Prepare dataset metadata for request."""
        from renku.command.schema.dataset import dump_dataset_as_jsonld

        jsonld = dump_dataset_as_jsonld(self.dataset)
        jsonld["upload_type"] = "dataset"
        return jsonld

    def export(self, **kwargs):
        """Execute entire export process."""
        # Step 1. Create new deposition
        deposition = ZenodoDeposition(exporter=self)

        # Step 2. Attach metadata to deposition
        deposition.attach_metadata(self.dataset, self._tag)

        # Step 3. Upload all files to created deposition
        with communication.progress("Uploading files ...", total=len(self.dataset.files)) as progressbar:
            for file in self.dataset.files:
                filepath = project_context.repository.copy_content_to_file(
                    path=file.entity.path, checksum=file.entity.checksum
                )
                deposition.upload_file(filepath, path_in_repo=file.entity.path)
                progressbar.update()

        # Step 4. Publish newly created deposition
        if self._publish:
            deposition.publish_deposition()
            return deposition.published_at

        return deposition.deposit_at


class ZenodoDeposition:
    """Zenodo record for a deposit."""

    def __init__(self, exporter, id=None):
        self.exporter = exporter
        self.id = id

        response = self.new_deposition()
        self.id = response.json()["id"]

    @property
    def publish_url(self):
        """Returns publish URL."""
        url = urllib.parse.urljoin(
            self.exporter.zenodo_url,
            posixpath.join(ZENODO_API_PATH, ZENODO_DEPOSIT_PATH, ZENODO_PUBLISH_ACTION_PATH.format(self.id)),
        )

        return url

    @property
    def attach_metadata_url(self):
        """Return URL for attaching metadata."""
        url = urllib.parse.urljoin(
            self.exporter.zenodo_url,
            posixpath.join(ZENODO_API_PATH, ZENODO_DEPOSIT_PATH, ZENODO_METADATA_URL.format(self.id)),
        )
        return url

    @property
    def upload_file_url(self):
        """Return URL for uploading file."""
        url = urllib.parse.urljoin(
            self.exporter.zenodo_url,
            posixpath.join(ZENODO_API_PATH, ZENODO_DEPOSIT_PATH, ZENODO_FILES_URL.format(self.id)),
        )
        return url

    @property
    def new_deposit_url(self):
        """Return URL for creating new deposit."""
        url = urllib.parse.urljoin(
            self.exporter.zenodo_url,
            posixpath.join(ZENODO_API_PATH, ZENODO_DEPOSIT_PATH, ZENODO_NEW_DEPOSIT_URL),
        )
        return url

    @property
    def published_at(self):
        """Return published at URL."""
        url = urllib.parse.urljoin(self.exporter.zenodo_url, posixpath.join(ZENODO_PUBLISH_PATH, str(self.id)))
        return url

    @property
    def deposit_at(self):
        """Return deposit at URL."""
        url = urllib.parse.urljoin(self.exporter.zenodo_url, posixpath.join(ZENODO_DEPOSIT_PATH, str(self.id)))
        return url

    def new_deposition(self):
        """Create new deposition on Zenodo."""
        from renku.core.util import requests

        response = requests.post(
            url=self.new_deposit_url, params=self.exporter.default_params, json={}, headers=self.exporter.HEADERS
        )
        self._check_response(response)

        return response

    def upload_file(self, filepath, path_in_repo):
        """Upload and attach a file to existing deposition on Zenodo."""
        from renku.core.util import requests

        request_payload = {"filename": Path(path_in_repo).name}
        file = {"file": (Path(path_in_repo).name, open(str(filepath), "rb"))}
        response = requests.post(
            url=self.upload_file_url, params=self.exporter.default_params, data=request_payload, files=file
        )
        self._check_response(response)

        return response

    def attach_metadata(self, dataset, tag):
        """Attach metadata to deposition on Zenodo."""
        from renku.core.util import requests

        request_payload = {
            "metadata": {
                "title": dataset.name,
                "upload_type": "dataset",
                "description": dataset.description if dataset.description else None,
                "creators": [
                    {"name": creator.name, "affiliation": creator.affiliation if creator.affiliation else None}
                    for creator in dataset.creators
                ],
                "keywords": dataset.keywords,
            }
        }

        version = tag.name if tag else dataset.version

        if version:
            request_payload["metadata"]["version"] = version

        response = requests.put(
            url=self.attach_metadata_url,
            params=self.exporter.default_params,
            data=json.dumps(request_payload),
            headers=self.exporter.HEADERS,
        )
        self._check_response(response)

        return response

    def publish_deposition(self):
        """Publish existing deposition."""
        from renku.core.util import requests

        response = requests.post(
            url=self.publish_url, params=self.exporter.default_params, headers=self.exporter.HEADERS
        )
        self._check_response(response)

        return response

    @staticmethod
    def _check_response(response):
        from renku.core.util import requests

        try:
            requests.check_response(response=response)
        except errors.RequestError:
            if response.status_code == 400:
                err_response = response.json()
                if "errors" in err_response:
                    messages = [
                        '"{}" failed with "{}"'.format(err["field"], ", ".join(err["messages"]))
                        for err in err_response["errors"]
                    ]
                elif "message" in err_response:
                    messages = [err_response["message"]]
                else:
                    messages = [response.text()]

                raise errors.ExportError(
                    "\n" + "\n".join(messages) + "\nSee `renku dataset edit -h` for details on how to edit" " metadata"
                )
            else:
                print(response.status_code)
                raise errors.ExportError(response.content)


def _make_request(uri, accept: str = "application/json"):
    """Execute network request."""
    record_id = ZenodoProvider.get_record_id(uri)
    url = make_records_url(record_id, uri=uri)

    return make_request(url=url, accept=accept)


def make_records_url(record_id, uri: str):
    """Create URL to access record by ID.

    Args:
        record_id:  The id of the record.

    Returns:
        str: Full URL for the record.
    """
    url = ZENODO_SANDBOX_URL if "sandbox.zenodo.org" in uri.lower() else ZENODO_BASE_URL

    return urllib.parse.urljoin(url, posixpath.join(ZENODO_API_PATH, "records", record_id))



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku workflow update."""

from pathlib import Path
from typing import List, Optional

from pydantic import ConfigDict, validate_call

from renku.core import errors
from renku.core.errors import ParameterError
from renku.core.util.os import get_relative_paths
from renku.core.workflow.activity import (
    get_all_modified_and_deleted_activities_and_entities,
    get_downstream_generating_activities,
    is_activity_valid,
    sort_activities,
)
from renku.core.workflow.execute import execute_workflow_graph
from renku.core.workflow.model.concrete_execution_graph import ExecutionGraph
from renku.domain_model.project_context import project_context


@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def update(
    update_all: bool,
    dry_run: bool,
    ignore_deleted: bool,
    provider: str,
    config: Optional[str],
    paths: Optional[List[str]] = None,
):
    """Update stale generated outputs."""
    if paths and update_all:
        raise ParameterError("Cannot use PATHS and --all/-a at the same time.")

    paths = paths or []
    paths = get_relative_paths(base=project_context.path, paths=[Path.cwd() / p for p in paths])

    modified, _, _ = get_all_modified_and_deleted_activities_and_entities(project_context.repository)
    modified_activities = {a for a, _ in modified if not a.deleted and is_activity_valid(a)}
    modified_paths = {e.path for _, e in modified}

    activities = get_downstream_generating_activities(
        starting_activities=modified_activities,
        paths=paths,
        ignore_deleted=ignore_deleted,
        project_path=project_context.path,
    )

    if len(activities) == 0:
        raise errors.NothingToExecuteError()

    # NOTE: When updating we only want to eliminate activities that are overridden, not their parents
    activities = sort_activities(activities, remove_overridden_parents=False)
    if dry_run:
        return activities, modified_paths

    graph = ExecutionGraph([a.plan_with_values for a in activities], virtual_links=True)
    execute_workflow_graph(dag=graph.workflow_graph, provider=provider, config=config)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service templates view."""
from flask import request

from renku.ui.service.config import SERVICE_PREFIX
from renku.ui.service.controllers.templates_create_project import TemplatesCreateProjectCtrl
from renku.ui.service.controllers.templates_read_manifest import TemplatesReadManifestCtrl
from renku.ui.service.views.api_versions import VERSIONS_FROM_V2_2, VersionedBlueprint
from renku.ui.service.views.decorators import accepts_json, requires_cache, requires_identity
from renku.ui.service.views.error_handlers import (
    handle_common_except,
    handle_templates_create_errors,
    handle_templates_read_errors,
)
from renku.ui.service.views.v1.templates import add_v1_specific_template_endpoints, add_v2_specific_template_endpoints

TEMPLATES_BLUEPRINT_TAG = "templates"
templates_blueprint = VersionedBlueprint(TEMPLATES_BLUEPRINT_TAG, __name__, url_prefix=SERVICE_PREFIX)


@templates_blueprint.route(
    "/templates.read_manifest", methods=["GET"], provide_automatic_options=False, versions=VERSIONS_FROM_V2_2
)
@handle_common_except
@handle_templates_read_errors
@requires_cache
@requires_identity
def read_manifest_from_template(user_data, cache):
    """
    Read templates from the manifest file of a template repository.

    ---
    get:
      description: Read templates from the manifest file of a template repository.
      parameters:
        - in: query
          name: url
          required: true
          schema:
            type: string
        - in: query
          name: ref
          schema:
            type: string
        - in: query
          name: depth
          schema:
            type: string
      responses:
        200:
          description: Listing of templates in the repository.
          content:
            application/json:
              schema: ManifestTemplatesResponseRPC
      tags:
        - templates
    """
    return TemplatesReadManifestCtrl(cache, user_data, dict(request.args)).to_response()


@templates_blueprint.route(
    "/templates.create_project", methods=["POST"], provide_automatic_options=False, versions=VERSIONS_FROM_V2_2
)
@handle_common_except
@handle_templates_create_errors
@accepts_json
@requires_cache
@requires_identity
def create_project_from_template(user_data, cache):
    """
    Create a new project starting using a remote template.

    ---
    post:
      description: Create a new project using a remote template.
      requestBody:
        content:
          application/json:
            schema: ProjectTemplateRequest
      responses:
        200:
          description: Details of the created project.
          content:
            application/json:
              schema: ProjectTemplateResponseRPC
      tags:
        - templates
    """
    return TemplatesCreateProjectCtrl(cache, user_data, dict(request.json)).to_response()  # type: ignore


templates_blueprint = add_v1_specific_template_endpoints(templates_blueprint)
templates_blueprint = add_v2_specific_template_endpoints(templates_blueprint)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku API."""

from renku.ui.api.graph.rdf import RDFGraph
from renku.ui.api.models.activity import Activity
from renku.ui.api.models.dataset import Dataset
from renku.ui.api.models.parameter import Input, Link, Mapping, Output, Parameter
from renku.ui.api.models.plan import CompositePlan, Plan
from renku.ui.api.models.project import Project

__all__ = (
    "Activity",
    "CompositePlan",
    "Dataset",
    "Input",
    "Link",
    "Mapping",
    "Output",
    "Parameter",
    "Plan",
    "Project",
    "RDFGraph",
)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service apispec views."""
from apispec import APISpec, yaml_utils
from apispec_oneofschema import MarshmallowPlugin
from apispec_webframeworks.flask import FlaskPlugin
from flask import Blueprint, current_app, jsonify
from flask.views import MethodView

from renku.ui.service.config import (
    API_VERSION,
    OIDC_URL,
    OPENAPI_VERSION,
    SERVICE_API_BASE_PATH,
    SERVICE_NAME,
    SERVICE_PREFIX,
)

apispec_blueprint = Blueprint("apispec", __name__, url_prefix=SERVICE_PREFIX)

# security schemes
oidc_scheme = {"type": "openIdConnect", "openIdConnectUrl": OIDC_URL}
jwt_scheme = {"type": "apiKey", "name": "Renku-User", "in": "header"}
gitlab_token_scheme = {"type": "apiKey", "name": "Authorization", "in": "header"}

TOP_LEVEL_DESCRIPTION = """
This is the API specification of the renku core service.

## Responses

Loosely following the JSON-RPC 2.0 Specification, the methods all return with
HTTP code 200 and include a [response
object](https://www.jsonrpc.org/specification#response_object) may contain
either a `result` or an `error` object. If the call succeeds, the returned
`result` follows the schema documented in the individual methods. In the case of
an error, the [`error`
object](https://www.jsonrpc.org/specification#error_object), contains a code and
two messages describing the nature of the error. One targets the user, while the
other targets the developers. Instead of following the [standard JSON-RPC
response codes](https://www.jsonrpc.org/specification#error_object), we define
our own application-specific codes. Those can be found in the definition of the
error service classes.
"""


class MultiURLFlaskPlugin(FlaskPlugin):
    """FlaskPlugin extension that supports multiple URLs per endpoint."""

    def path_helper(self, path, operations, *, view, app=None, **kwargs):
        """Path helper that allows passing a Flask view function."""
        rule = self._rule_for_view(view, app=app)
        operations.update(yaml_utils.load_operations_from_docstring(view.__doc__))
        if hasattr(view, "view_class") and issubclass(view.view_class, MethodView):
            for method in view.methods:
                if method in rule.methods:
                    method_name = method.lower()
                    method = getattr(view.view_class, method_name)
                    operations[method_name] = yaml_utils.load_yaml_from_docstring(method.__doc__)
        return self.flaskpath2openapi(path)


spec = APISpec(
    title=SERVICE_NAME,
    openapi_version=OPENAPI_VERSION,
    version=API_VERSION,
    plugins=[MultiURLFlaskPlugin(), MarshmallowPlugin()],
    servers=[{"url": SERVICE_API_BASE_PATH}, {"url": f"/ui-server{SERVICE_API_BASE_PATH}"}],
    security=[{"oidc": []}, {"JWT": [], "gitlab-token": []}],
    info={"description": TOP_LEVEL_DESCRIPTION},
)

spec.components.security_scheme("oidc", oidc_scheme)
spec.components.security_scheme("jwt", jwt_scheme)
spec.components.security_scheme("gitlab-token", gitlab_token_scheme)


@apispec_blueprint.route("/spec.json")
def openapi():
    """Return the OpenAPI spec for this service."""
    return jsonify(get_apispec(current_app).to_dict())


def get_apispec(app):
    """Return the apispec."""
    for rule in current_app.url_map.iter_rules():
        spec.path(path=rule.rule, view=app.view_functions[rule.endpoint])
    return spec
